"""Constants for the victron venus OS client."""

from enum import Enum

TOPIC_INSTALLATION_ID = "N/+/system/0/Serial"


class MessageType(Enum):
    """Type of MQTT message."""

    ATTRIBUTE = "attribute"
    METRIC = "metric"


class MetricNature(Enum):
    """Nature of the metric."""

    NONE = "none"
    INSTANTANEOUS = "instantaneous"
    CUMULATIVE = "cumulative"
    DELTA = "delta"


class MetricType(Enum):
    """The type of metric."""

    NONE = "none"
    POWER = "power"
    ENERGY = "energy"
    VOLTAGE = "voltage"
    CURRENT = "current"
    TEMPERATURE = "temperature"
    PRESSURE = "pressure"
    FREQUENCY = "frequency"
    TIME = "time"
    PERCENTAGE = "percentage"
    ELECTRIC_STORAGE_CAPACITY = "electric_storage_capacity"


class DeviceType(Enum):
    """Type of device."""

    ANY = "any"
    SYSTEM = "system"
    SOLAR_CHARGER = "solar_charger"
    INVERTER = "inverter"
    BATTERY = "battery"
    GRID = "grid"


PLACEHOLDER_PHASE = "{phase}"

DEFAULT_HOST = "venus.local."
DEFAULT_PORT = 1883



"""
Maps all the MQTT topics to either attributes or metrics.
"""

from victronvenusclient._unwrappers import unwrap_float, unwrap_int, unwrap_int_default_0, unwrap_string
from victronvenusclient.constants import DeviceType, MessageType, MetricNature, MetricType
from victronvenusclient.data_classes import TopicDescriptor

topic_map = {
    # generic device attributes
    "N/+/+/+/ProductName": TopicDescriptor(
        message_type=MessageType.ATTRIBUTE,
        short_id="model",
        unit_of_measurement=None,
        metric_type=None,
        metric_nature=None,
        precision=None,
        unwrapper=unwrap_string,
    ),
    "N/+/+/+/Serial": TopicDescriptor(
        message_type=MessageType.ATTRIBUTE,
        short_id="serial_number",
        unit_of_measurement=None,
        metric_type=MetricType.NONE,
        metric_nature=None,
        precision=None,
        unwrapper=unwrap_string,
    ),
    # inverter hides its serial number away in a different topic
    "N/+/vebus/+/Devices/0/SerialNumber": TopicDescriptor(
        message_type=MessageType.ATTRIBUTE,
        short_id="serial_number",
        unit_of_measurement=None,
        metric_type=MetricType.NONE,
        metric_nature=None,
        precision=None,
        unwrapper=unwrap_string,
    ),
    "N/+/+/+/Manufacturer": TopicDescriptor(
        message_type=MessageType.ATTRIBUTE,
        short_id="manufacturer",
        unit_of_measurement=None,
        metric_type=MetricType.NONE,
        metric_nature=None,
        precision=None,
        unwrapper=unwrap_string,
    ),
    "N/+/+/+/ProductId": TopicDescriptor(
        message_type=MessageType.ATTRIBUTE,
        short_id="victron_productid",
        unit_of_measurement=None,
        metric_type=MetricType.NONE,
        metric_nature=None,
        precision=0,
        unwrapper=unwrap_int,
    ),
    "N/+/+/+/FirmwareVersion": TopicDescriptor(
        message_type=MessageType.ATTRIBUTE,
        short_id="firmware_version",
        unit_of_measurement=None,
        metric_type=MetricType.NONE,
        metric_nature=None,
        precision=None,
        unwrapper=unwrap_string,
    ),
    # grid
    "N/+/system/+/Ac/Grid/NumberOfPhases": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="system_grid_phases",  # system attribute
        unit_of_measurement=None,
        metric_type=MetricType.NONE,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.GRID,
        precision=0,
        unwrapper=unwrap_int_default_0,
    ),
    # individual grid phases
    "N/+/grid/+/Ac/+/Voltage": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="grid_voltage_{phase}",
        unit_of_measurement="V",
        metric_type=MetricType.VOLTAGE,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.GRID,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/grid/+/Ac/+/Current": TopicDescriptor(
        message_type="sensor",
        short_id="grid_current_{phase}",
        unit_of_measurement="A",
        metric_type=MetricType.CURRENT,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.GRID,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/grid/+/Ac/+/Power": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="grid_power_{phase}",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.GRID,
        precision=1,
        unwrapper=unwrap_float,
    ),
    # total grid
    "N/+/grid/+/Ac/Voltage": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="grid_voltage",
        unit_of_measurement="V",
        metric_type=MetricType.VOLTAGE,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.GRID,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/grid/+/Ac/Current": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="grid_current",
        unit_of_measurement="A",
        metric_type=MetricType.CURRENT,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.GRID,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/grid/+/Ac/Power": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="grid_power",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.GRID,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/grid/+/Ac/Energy/Forward": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="grid_energy_forward",
        unit_of_measurement="kWh",
        metric_type=MetricType.ENERGY,
        metric_nature=MetricNature.CUMULATIVE,
        device_type=DeviceType.GRID,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/grid/+/Ac/Energy/Reverse": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="grid_energy_reverse",
        unit_of_measurement="kWh",
        metric_type=MetricType.ENERGY,
        metric_nature=MetricNature.CUMULATIVE,
        device_type=DeviceType.GRID,
        precision=1,
        unwrapper=unwrap_float,
    ),
    #  cspell:ignore solarcharger MPPT
    # solar / MPPT  """ cspell:ignore  """
    "N/+/solarcharger/+/Dc/0/Voltage": TopicDescriptor(  # """ cspell:ignore  """
        message_type=MessageType.METRIC,
        short_id="solar_voltage",
        unit_of_measurement="V",
        metric_type=MetricType.VOLTAGE,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.SOLAR_CHARGER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/solarcharger/+/Dc/0/Current": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="solar_current",
        unit_of_measurement="A",
        metric_type=MetricType.CURRENT,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.SOLAR_CHARGER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/solarcharger/+/Yield/Power": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="solar_power",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.SOLAR_CHARGER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/solarcharger/+/Yield/User": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="solar_yield",
        unit_of_measurement="kWh",
        metric_type=MetricType.ENERGY,
        metric_nature=MetricNature.CUMULATIVE,
        device_type=DeviceType.SOLAR_CHARGER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/solarcharger/+/History/Daily/0/MaxPower": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="solar_max_power",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.SOLAR_CHARGER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    # batteries
    "N/+/battery/+/Dc/0/Voltage": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_voltage",
        unit_of_measurement="V",
        metric_type=MetricType.VOLTAGE,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/battery/+/Dc/0/Current": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_current",
        unit_of_measurement="A",
        metric_type=MetricType.CURRENT,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/battery/+/Dc/0/Power": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_power",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/battery/+/Dc/0/Temperature": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_temperature",
        unit_of_measurement="°C",
        metric_type=MetricType.TEMPERATURE,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/battery/+/History/DischargedEnergy": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_discharged_energy",
        unit_of_measurement="kWh",
        metric_type=MetricType.ENERGY,
        metric_nature=MetricNature.CUMULATIVE,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/battery/+/History/ChargedEnergy": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_charged_energy",
        unit_of_measurement="kWh",
        metric_type=MetricType.ENERGY,
        metric_nature=MetricNature.CUMULATIVE,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/battery/+/Capacity": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_capacity",
        unit_of_measurement="Ah",
        metric_type=MetricType.ELECTRIC_STORAGE_CAPACITY,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/battery/+/InstalledCapacity": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_installed_capacity",
        unit_of_measurement="Ah",
        metric_type=MetricType.ELECTRIC_STORAGE_CAPACITY,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/battery/+/Soc": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="battery_soc",
        unit_of_measurement="%",
        metric_type=MetricType.PERCENTAGE,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.BATTERY,
        precision=1,
        unwrapper=unwrap_float,
    ),
    # inverter
    "N/+/vebus/+/Ac/ActiveIn/+/P": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="inverter_input_power_{phase}",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.INVERTER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/vebus/+/Ac/ActiveIn/+/F": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="inverter_input_frequency_{phase}",
        unit_of_measurement="Hz",
        metric_type=MetricType.FREQUENCY,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.INVERTER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/vebus/+/Ac/ActiveIn/+/S": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="inverter_input_apparent_power_{phase}",
        unit_of_measurement="VA",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.INVERTER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/vebus/+/Ac/Out/+/P": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="inverter_output_power_{phase}",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.INVERTER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/vebus/+/Ac/Out/+/F": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="inverter_output_frequency_{phase}",
        unit_of_measurement="Hz",
        metric_type=MetricType.FREQUENCY,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.INVERTER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/vebus/+/Ac/Out/+/S": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="inverter_output_apparent_power_{phase}",
        unit_of_measurement="VA",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.INVERTER,
        precision=1,
        unwrapper=unwrap_float,
    ),
    # integrated system. Note that this might not be currently accurate for all systems
    #  as there are different physical configurations
    # and don't have access to any other for testing or feedback.
    "N/+/system/+/Ac/ConsumptionOnOutput/+/Power": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="system_critical_loads_{phase}",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.SYSTEM,
        precision=1,
        unwrapper=unwrap_float,
    ),
    "N/+/system/+/Ac/ConsumptionOnInput/+/Power": TopicDescriptor(
        message_type=MessageType.METRIC,
        short_id="system_ac_loads_{phase}",
        unit_of_measurement="W",
        metric_type=MetricType.POWER,
        metric_nature=MetricNature.INSTANTANEOUS,
        device_type=DeviceType.SYSTEM,
        precision=1,
        unwrapper=unwrap_float,
    ),
}



"""A simple utility to dump the entire MQTT state of a Venus OS device."""

import argparse
import json
import asyncio
from ..hub import Hub


def setup_arguments():
    """Setup the arguments for the script."""
    parser = argparse.ArgumentParser(description="Dump MQTT messages from a Venus OS hub")
    parser.add_argument("--host", default="venus.local.", help="Hostname of the Venus OS hub")
    parser.add_argument("--username", help="Username to use for the MQTT connection")
    parser.add_argument("--password", help="Password to use for the MQTT connection")
    parser.add_argument("--port", type=int, default=1883, help="Port of the Venus OS hub")
    parser.add_argument("--use-ssl", action="store_true", help="Use SSL for the connection")
    return parser.parse_args()


async def async_main(args):
    """Main function."""
    hub = Hub(host=args.host, port=args.port, username=args.username, password=args.password, use_ssl=args.use_ssl)

    await hub.connect()
    output = await hub.create_full_raw_snapshot()
    await hub.disconnect()
    print(json.dumps(output, indent=4))


def main():
    """Wrapper for async main function."""
    args = setup_arguments()
    asyncio.run(async_main(args))


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024-present Johan du Plessis https://github.com/johanslab
#
# SPDX-License-Identifier: MIT
"""Configuration for tests, mostly Victron Venus device connection settings.

Can be overridden by environment variables:
VENUS_TEST_HOST - The host or IP address of the Venus device. Default "venus.local."
VENUS_TEST_PORT - The port of the Venus device. Default 1883
VENUS_TEST_USERNAME - The username for the Venus device. Default None
VENUS_TEST_PASSWORD - The password for the Venus device. Default None
VENUS_TEST_USE_SSL - Whether to use SSL for the connection. Default False

"""

import os
import pytest


@pytest.fixture
def config_host():
    return os.getenv("VENUS_TEST_HOST", "venus.local.")


@pytest.fixture
def config_port():
    return int(os.getenv("VENUS_TEST_PORT", "1883"))


@pytest.fixture
def config_username():
    return os.getenv("VENUS_TEST_USERNAME", None)


@pytest.fixture
def config_password():
    return os.getenv("VENUS_TEST_PASSWORD", None)


@pytest.fixture
def config_use_ssl() -> bool:
    use_ssl_str = os.getenv("VENUS_TEST_USE_SSL", "False")
    return use_ssl_str.lower() in ["true", "1", "t", "y", "yes", "on"]



# SPDX-FileCopyrightText: 2024-present Johan du Plessis https://github.com/johanslab
#
# SPDX-License-Identifier: MIT
"""Tests basic connectivity functionality. Does require a running Venus OS instance to connect to."""

import pytest
import victronvenusclient  # pylint: disable=import-error


@pytest.mark.asyncio
async def test_connect(config_host, config_port, config_username, config_password, config_use_ssl):
    """Tests whether the client can connect to a Venus device. Disconnects after passing the test."""
    hub = victronvenusclient.Hub(config_host, config_port, config_username, config_password, config_use_ssl)
    await hub.connect()
    assert hub.connected
    await hub.disconnect()


@pytest.mark.asyncio
async def test_verify_connection(config_host, config_port, config_username, config_password, config_use_ssl):
    """
    Tests whether the client can connect to a Venus device and verify the connection details by
    checking whether a serial number could be obtained.
    """
    hub = victronvenusclient.Hub(config_host, config_port, config_username, config_password, config_use_ssl)
    serial = await hub.verify_connection_details()
    assert len(serial) > 0



# SPDX-FileCopyrightText: 2024-present Johan du Plessis https://github.com/johanslab
#
# SPDX-License-Identifier: MIT

"""Tests that devices and metrics can be enumerated."""

import pytest
import victronvenusclient  # pylint: disable=import-error


@pytest.mark.asyncio
async def test_devices_and_metrics(config_host, config_port, config_username, config_password, config_use_ssl):
    hub = victronvenusclient.Hub(config_host, config_port, config_username, config_password, config_use_ssl)
    await hub.connect()
    await hub.initialize_devices_and_metrics()

    assert len(hub.devices) > 0

    for device in hub.devices:
        assert len(device.metrics) > 0
        assert device.device_type is not None
        assert device.device_type != victronvenusclient.DeviceType.ANY

        for metric in device.metrics:
            assert len(metric.short_id) > 0

    await hub.disconnect()



"""Logic for handling Victron devices, and routing updates to the appropriate metrics."""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

from victronvenusclient.constants import DeviceType, PLACEHOLDER_PHASE, MessageType
from victronvenusclient.metric import Metric

if TYPE_CHECKING:
    from victronvenusclient.data_classes import ParsedTopic, TopicDescriptor

_LOGGER = getLogger(__name__)


class Device:
    """Class to represent a Victron device."""

    def __init__(
        self,
        unique_id: str,
        descriptor: TopicDescriptor,
        installation_id: str,
        native_device_type: str,
        device_id: str,
    ) -> None:
        """Initialize."""
        self._descriptor = descriptor
        self._unique_id = unique_id
        self._metrics: dict[str, Metric] = {}
        self._root_device_name: str = None
        self._native_device_type = native_device_type
        self._device_type = descriptor.device_type
        self._device_id = device_id
        self._installation_id = installation_id
        self._device_name = ""
        self._model = ""
        self._manufacturer = ""
        self._serial_number = ""
        self._firmware_version = ""

    def _set_device_property_from_topic(
        self,
        parsed_topic: ParsedTopic,
        topic_desc: TopicDescriptor,
        payload: str,  # noqa: ARG002 pylint: disable=unused-argument
    ) -> None:
        """Set a device property from a topic."""
        short_id = topic_desc.short_id
        if topic_desc.unwrapper is not None:
            payload = str(topic_desc.unwrapper(payload))

        if payload is None:
            return

        if payload == "None":
            return

        if len(payload) == 0:
            return

        if short_id == "victron_productid":
            return  # ignore for now

        if short_id == "model":
            self._model = payload
        elif short_id == "serial_number":
            self._serial_number = payload
        elif short_id == "manufacturer":
            self._manufacturer = payload
        elif short_id == "firmware_version":
            self._firmware_version = payload
        else:
            _LOGGER.warning("Unhandled device property %s for %s", short_id, self.unique_id)

        # if we get a model message and we don't have a name yet, we use the model as name

        if short_id == "model" and self._device_name == "":
            self._device_name = payload

    async def handle_message(self, parsed_topic: ParsedTopic, topic_desc: TopicDescriptor, payload: str) -> None:
        """Handle a message."""

        # if we created the device on a generic topic we need to fix the device type as soon
        #  as we get a more specific topic
        if self.device_type == DeviceType.ANY and topic_desc.device_type != DeviceType.ANY:
            self._device_type = topic_desc.device_type

        if topic_desc.message_type == MessageType.ATTRIBUTE:
            self._set_device_property_from_topic(parsed_topic, topic_desc, payload)
        elif topic_desc.message_type == MessageType.METRIC:
            value = payload
            if topic_desc.unwrapper is not None:
                value = topic_desc.unwrapper(payload)
            if value is None:
                return  # don't try to create or update metric if we don't have valid values for it.

            short_id = topic_desc.short_id
            if PLACEHOLDER_PHASE in short_id:
                short_id = short_id.replace(PLACEHOLDER_PHASE, parsed_topic.phase)
            metric_id = f"{self.unique_id}_{short_id}"

            metric = self._get_or_create_metric(metric_id, short_id, parsed_topic, topic_desc, payload)

            await metric.handle_message(parsed_topic, topic_desc, value)

    def _get_or_create_metric(
        self, metric_id: str, short_id: str, parsed_topic: ParsedTopic, topic_desc: TopicDescriptor, payload: str
    ) -> Metric:
        """Get or create a metric."""
        metric = self._metrics.get(metric_id)
        if metric is None:
            metric = Metric(metric_id, topic_desc, parsed_topic, payload)
            self._metrics[metric_id] = metric
            setattr(self, short_id, metric)

        return metric

    def set_root_device_name(self, name: str) -> None:
        """Set the name of the root device."""
        self._root_device_name = name
        self._device_name = name

    def get_metric_from_unique_id(self, unique_id: str) -> Metric:
        """Get a metric from a unique id."""
        return self._metrics.get(unique_id)

    @property
    def metrics(self) -> list[Metric]:
        """Returns the list of metrics on this device."""
        return list(self._metrics.values())

    @property
    def unique_id(self) -> str:
        """Return the unique id of the device."""
        return self._unique_id

    @property
    def name(self) -> str:
        """Return the name of the device."""
        return self._device_name

    @property
    def model(self) -> str:
        """Return the model of the device."""
        return self._model

    @property
    def manufacturer(self) -> str:
        """Return the manufacturer of the device."""
        return self._manufacturer

    @property
    def serial_number(self) -> str:
        """Return the serial number of the device."""
        return self._serial_number

    @property
    def device_type(self) -> DeviceType:
        """Return the device type."""
        return self._device_type

    @property
    def native_device_type(self) -> str:
        """Return the native device type."""
        return self._native_device_type

    @property
    def firmware_version(self) -> str:
        """Return the firmware version of the device."""
        return self._firmware_version



"""
Support for Victron Venus sensors. The sensor itself has no logic,
 it simply receives messages and updates its state.
"""

from __future__ import annotations

from collections.abc import Callable

from victronvenusclient.constants import PLACEHOLDER_PHASE
from victronvenusclient.data_classes import ParsedTopic, TopicDescriptor


class Metric:
    """Representation of a Victron Venus sensor."""

    def __init__(self, unique_id: str, descriptor: TopicDescriptor, parsed_topic: ParsedTopic, value) -> None:
        """Initialize the sensor."""

        self._descriptor = descriptor
        self._unique_id = unique_id
        self._value = value
        self._generic_short_id = descriptor.short_id.replace(PLACEHOLDER_PHASE, "lx")

        self._phase = parsed_topic.phase
        self._on_update: Callable = None

        if parsed_topic.phase is not None:
            self._short_id = descriptor.short_id.replace(PLACEHOLDER_PHASE, parsed_topic.phase)

    def __str__(self) -> str:
        """Return the string representation of the metric."""
        return self.formatted_value

    @property
    def formatted_value(self) -> str:
        """Returns the formatted value of the metric."""
        if self._value is None:
            return ""

        try:
            fvalue = float(self._value)  # Attempt to convert the object to a float
            if self._descriptor.unit_of_measurement is None:
                return f"{fvalue:.{self._descriptor.precision}f}"
            else:
                return f"{fvalue:.{self._descriptor.precision}f} {self._descriptor.unit_of_measurement}"
        except (ValueError, TypeError):  # Handle cases where conversion fails
            return str(self._value)

    @property
    def value(self):
        """Returns the value of the metric."""
        return self._value

    @property
    def short_id(self) -> str:
        """Returns the short id of the metric."""
        return self._short_id

    @property
    def generic_short_id(self) -> str:
        """Returns the generic short id of the metric."""
        return self._generic_short_id

    @property
    def phase(self) -> str:
        """Returns the phase of the metric if referring to a specific AC phase."""
        return self._phase

    @property
    def unit_of_measurement(self) -> str:
        """Returns the unit of measurement of the metric."""
        return self._descriptor.unit_of_measurement

    @property
    def metric_type(self):
        """Returns the metric type."""
        return self._descriptor.metric_type

    @property
    def metric_nature(self):
        """Returns the metric nature."""
        return self._descriptor.metric_nature

    @property
    def device_type(self):
        """Returns the device type."""
        return self._descriptor.device_type

    @property
    def precision(self):
        """Returns the precision of the metric."""
        return self._descriptor.precision

    @property
    def unique_id(self) -> str:
        """Return the unique id of the metric."""
        return self._unique_id

    @property
    def on_update(self) -> Callable:
        """Returns the on_update callback."""
        return self._on_update

    @on_update.setter
    def on_update(self, value: Callable):
        """Sets the on_update callback."""
        self._on_update = value

    async def handle_message(self, parsed_topic, topic_desc, value):  # noqa: ARG002 pylint: disable=unused-argument
        """Handle a message."""
        self._value = value
        if self.on_update is not None:
            self._on_update(self)



"""Functions to unwrap the data from the JSON string."""

import json


def unwrap_int(json_str):
    """Unwrap an integer value from a JSON string."""
    try:
        data = json.loads(json_str)
        return int(data["value"])
    except (json.JSONDecodeError, KeyError, ValueError, TypeError):
        return None


def unwrap_int_default_0(json_str):
    """Unwrap an integer value from a JSON string, defaulting to 0."""
    try:
        data = json.loads(json_str)
        if data["value"] is None:
            return 0
        return int(data["value"])
    except (json.JSONDecodeError, KeyError, ValueError, TypeError):
        return 0


def unwrap_float(json_str):
    """Unwrap a float value from a JSON string."""
    try:
        data = json.loads(json_str)
        if data["value"] is None:
            return 0.0
        return float(data["value"])
    except (json.JSONDecodeError, KeyError, ValueError, TypeError):
        return None


def unwrap_string(json_str):
    """Unwrap a string value from a JSON string."""
    try:
        data = json.loads(json_str)
        if data["value"] is None:
            return None
        return str(data["value"])
    except (json.JSONDecodeError, KeyError, ValueError, TypeError):
        return None



"""A simple utility to illustrate how to use the Victron Venus Client library
to connect to a Venus OS device and display the metrics."""

import asyncio
import inspect
from logging import getLogger
import tkinter as tk
from tkinter import simpledialog
import tkinter.ttk as ttk
from victronvenusclient import Hub, Device, Metric
from victronvenusclient.constants import DEFAULT_HOST, DEFAULT_PORT


LOGGER = getLogger(__name__)

"""Dialog to prompt for connection information"""


class ConnectionDialog(simpledialog.Dialog):
    """Dialog to prompt for connection information."""

    def body(self, master):
        tk.Label(master, text="Server:").grid(row=0)
        tk.Label(master, text="Port:").grid(row=1)
        tk.Label(master, text="Username:").grid(row=2)
        tk.Label(master, text="Password:").grid(row=3)
        tk.Label(master, text="Use SSL:").grid(row=4)

        self.server_entry = tk.Entry(master)
        self.server_entry.insert(0, DEFAULT_HOST)

        self.port_entry = tk.Entry(master)
        self.port_entry.insert(0, str(DEFAULT_PORT))

        self.username_entry = tk.Entry(master)

        self.password_entry = tk.Entry(master, show="*")

        self.use_ssl_var = tk.BooleanVar()
        self.use_ssl_check = tk.Checkbutton(master, variable=self.use_ssl_var)

        self.server_entry.grid(row=0, column=1)
        self.port_entry.grid(row=1, column=1)
        self.username_entry.grid(row=2, column=1)
        self.password_entry.grid(row=3, column=1)
        self.use_ssl_check.grid(row=4, column=1)

        return self.server_entry

    def apply(self):
        server = self.server_entry.get()
        port = int(self.port_entry.get())
        username = self.username_entry.get()
        password = self.password_entry.get()
        use_ssl = self.use_ssl_var.get()
        self.result = (server, port, username, password, use_ssl)


class AttributeViewerDialog(simpledialog.Dialog):
    """Dialog to display the attributes of an object."""

    def __init__(self, parent, instance):
        self.instance = instance
        super().__init__(parent, title="Attribute Viewer")

    def body(self, master):
        row = 0
        for name, value in inspect.getmembers(type(self.instance), lambda v: isinstance(v, property)):
            try:
                value = getattr(self.instance, name)
                if isinstance(value, (Device, Metric, list, dict)):
                    continue
                if callable(value):
                    continue
                ttk.Label(master, text=f"{name}:").grid(row=row, column=0, sticky=tk.W, padx=5, pady=5)
                ttk.Label(master, text=str(value)).grid(row=row, column=1, sticky=tk.W, padx=5, pady=5)
                row += 1
            except AttributeError:
                continue

        return master

    def buttonbox(self):
        box = ttk.Frame(self)

        w = ttk.Button(box, text="OK", width=10, command=self.ok, default=tk.ACTIVE)
        w.pack(side=tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()


class MetricContainer:
    """A UI orientated container for a metric. Updates the item in the tree view when the metric changes."""

    def __init__(self, metric: Metric, tree_view: ttk.Treeview, parent_item: str):
        self._metric = metric
        self._metric.on_update = self._update
        self._tree_view = tree_view
        self._parent_item = parent_item

    def _update(self, metric: Metric):  # pylint: disable=unused-argument
        formatted = self._metric.formatted_value
        self._tree_view.item(self._parent_item, values=(formatted,))


class App:
    def __init__(self):
        self.root = tk.Tk()

        self.root.resizable(True, True)
        self.root.geometry("1024x768")

        self.root = self.root
        self.root.title("Victron Venus Client")

        self.toolbar = ttk.Frame(self.root)
        self.toolbar.pack(side=tk.TOP, fill=tk.X)

        self.connect_button = ttk.Button(self.toolbar, text="Connect", command=self._connect)
        self.connect_button.pack(side=tk.LEFT, padx=2, pady=2)

        self.disconnect_button = ttk.Button(
            self.toolbar, text="Disconnect", command=self._disconnect, state=tk.DISABLED
        )
        self.disconnect_button.pack(side=tk.LEFT, padx=2, pady=2)

        self.info_button = ttk.Button(self.toolbar, text="Info", command=self._info, state=tk.DISABLED)
        self.info_button.pack(side=tk.LEFT, padx=10, pady=2)

        self.tree = ttk.Treeview(self.root, selectmode="browse")
        self.tree.bind("<<TreeviewSelect>>", self._on_tree_select)
        self.tree.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        self.tree["columns"] = "value"
        self.tree.heading("#0", text="Item")
        self.tree.heading("value", text="Value")

        self._client = None
        self._metric_containers = []

        self.root.protocol("WM_DELETE_WINDOW", self._on_close)

        self._to_quit = False

    def update(self):
        self.root.update()

    def _on_close(self):
        self._to_quit = True

    @property
    def to_quit(self):
        return self._to_quit

    def _on_tree_select(self, event):  # pylint: disable=unused-argument
        if len(self.tree.selection()) == 0:  # nothing selected
            if str(self.info_button["state"]) == tk.NORMAL:
                self.info_button.config(state=tk.DISABLED)
        else:
            if str(self.info_button["state"]) == tk.DISABLED:
                self.info_button.config(state=tk.NORMAL)

    async def _async_connect(self, server: str, port: int, username: str, password: str, use_ssl: bool) -> bool:
        try:
            self._client = Hub(server, port, username, password, use_ssl)
            await self._client.connect()
            await self._client.initialize_devices_and_metrics()
            self._fill_tree()
            self.disconnect_button.config(state=tk.NORMAL)
        except Exception as e:  # pylint: disable=broad-except
            message = str(e)
            if message == "":
                message = type(e).__name__
            simpledialog.messagebox.showerror("Error", f"Error connecting: {message}")
            await self._client.disconnect()
            self.connect_button.config(state=tk.NORMAL)

    def _fill_tree(self):
        for device in self._client.devices:
            device_item = self.tree.insert(
                "",
                "end",
                text=device.model,
                values=(device.serial_number, ""),
                iid="D" + device.unique_id,
            )
            metrics = device.metrics
            metrics.sort(key=lambda x: (x.phase if len(x.phase) > 0 else "L0") + x.short_id)
            for metric in metrics:
                metric_item = self.tree.insert(
                    device_item,
                    "end",
                    text=metric.short_id,
                    values=(metric.formatted_value,),
                    iid="M" + metric.unique_id,
                )
                self._metric_containers.append(MetricContainer(metric, self.tree, metric_item))

    def _info(self):
        item = self.tree.selection()[0]
        unique_id = item[1:]
        if item[0] == "D":
            device = self._client.get_device_from_unique_id(unique_id)
            if device is not None:
                AttributeViewerDialog(self.root, device)
        elif item[0] == "M":
            metric = self._client.get_metric_from_unique_id(unique_id)
            if metric is not None:
                AttributeViewerDialog(self.root, metric)

    def _connect(self):
        self.connect_button.config(state=tk.DISABLED)

        if self._client is not None:
            self._disconnect()

        dialog = ConnectionDialog(self.root)
        if dialog.result:
            server, port, username, password, use_ssl = dialog.result
        else:
            self.connect_button.config(state=tk.NORMAL)
            return

        if server == "":
            server = DEFAULT_HOST
        if port == 0:
            port = DEFAULT_PORT
        if username == "":
            username = None
        if password == "":
            password = None

        asyncio.run_coroutine_threadsafe(
            self._async_connect(server, port, username, password, use_ssl),
            asyncio.get_event_loop(),
        )

    def _disconnect(self):
        self.disconnect_button.config(state=tk.DISABLED)
        self.info_button.config(state=tk.DISABLED)
        self.tree.delete(*self.tree.get_children())

        if self._client is not None:
            loop = asyncio.get_event_loop()
            asyncio.run_coroutine_threadsafe(self._client.disconnect(), loop)
            self._client = None

        self.connect_button.config(state=tk.NORMAL)


def main():
    app = App()

    async def main_loop():
        while not app.to_quit:
            app.update()
            await asyncio.sleep(0.01)

    asyncio.run(main_loop())


if __name__ == "__main__":
    main()



"""Module to communicate with the Venus OS MQTT Broker."""

import asyncio
import json
import logging

from gmqtt import Client as gmqttClient
from gmqtt.mqtt.handler import MQTTConnectError

from victronvenusclient._topic_map import topic_map
from victronvenusclient.constants import TOPIC_INSTALLATION_ID
from victronvenusclient.data_classes import ParsedTopic, TopicDescriptor
from victronvenusclient.device import Device
from victronvenusclient.metric import Metric


class Hub:
    """Class to communicate with the Venus OS hub."""

    _installation_id_event: asyncio.Event

    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        use_ssl: bool,
        installation_id: str = None,
        model_name: str = None,
        serial: str = "noserial",
    ) -> None:
        """Initialize."""
        self._model_name = model_name
        self.host = host
        self.username = username
        self.password = password
        self.serial = serial
        self.use_ssl = use_ssl
        self._client = None
        self.port = port
        self._installation_id = installation_id
        self.logger = logging.getLogger(__name__)
        self._devices: dict[str, Device] = {}
        self._first_refresh_event = asyncio.Event()
        self._installation_id_event = None
        self._snapshot = {}

    async def connect(self) -> None:
        """Connect to the hub."""
        self._client = gmqttClient("python-victron-venus")
        if self.username not in {None, ""}:
            self._client.set_auth_credentials(self.username, self.password)
        await self._client.connect(host=self.host, port=self.port, ssl=self.use_ssl)

    async def initialize_devices_and_metrics(self) -> None:
        """Initialize devices and all the metrics."""
        if self._installation_id is None:
            self._installation_id = await self._read_installation_id()
        await self._setup_subscriptions()
        await self.keep_alive()
        await self._wait_for_first_refresh()

    async def disconnect(self) -> None:
        """Disconnect from the hub."""
        if self._client is None:
            return
        if self._client.is_connected:
            await self._client.disconnect()
        self._client = None

    async def keep_alive(self) -> None:
        """Send a keep alive message to the hub. Updates will only be made to the metrics
        for the 60 seconds following this method call."""
        # cspell:disable-next-line
        keep_alive_topic = f"R/{self._installation_id}/keepalive"

        if self._client is not None:
            if self._client.is_connected:
                self._client.publish(keep_alive_topic, b"1")

    async def create_full_raw_snapshot(self) -> dict:
        """Create a full raw snapshot of the current state of the Venus OS device.
        Should not be used in conjunction with initialize_devices_and_metrics()."""
        self._snapshot = {}
        if self._installation_id is None:
            self._installation_id = await self._read_installation_id()
        self._client.on_message = self._on_snapshot_message
        self._client.subscribe("#")
        await self.keep_alive()
        await self._wait_for_first_refresh()
        return self._snapshot

    def _set_nested_dict_value(self, d: dict, keys: list[str], value: str) -> None:
        for key in keys[:-1]:
            d = d.setdefault(key, {})
        d[keys[-1]] = value

    async def _on_snapshot_message(
        self,
        client: gmqttClient,
        topic: str,
        payload: bytes,
        qos: int,
        retain: bool,  # pylint: disable=unused-argument
    ) -> None:
        topic_parts = topic.split("/")
        value = json.loads(payload.decode())
        self._set_nested_dict_value(self._snapshot, topic_parts, value)
        if "full_publish_completed" in topic:
            self._first_refresh_event.set()

    async def _on_installation_id_message(
        self,
        client: gmqttClient,
        topic: str,
        payload: bytes,
        qos: int,
        retain: bool,  # pylint: disable=unused-argument
    ) -> None:
        """Handle an incoming message from the hub."""

        topic_parts = topic.split("/")

        if len(topic_parts) != 5:
            return

        if topic_parts[2] == "system" and topic_parts[3] == "0" and topic_parts[4] == "Serial":
            payload_json = json.loads(payload.decode())
            self._installation_id = payload_json.get("value")
            self._installation_id_event.set()

    async def _read_installation_id(self) -> str:
        """Read the installation id for the Victron installation. Depends on no other
        subscriptions being active."""

        if self._client is None:
            raise ProgrammingError
        if not self._client.is_connected:
            raise NotConnectedError
        self._client.on_message = self._on_installation_id_message
        self._installation_id_event = asyncio.Event()
        self._client.subscribe(TOPIC_INSTALLATION_ID)
        await asyncio.wait_for(self._installation_id_event.wait(), timeout=60)
        self._client.unsubscribe(TOPIC_INSTALLATION_ID)
        return str(self.installation_id)

    async def verify_connection_details(self) -> str:
        """Verify the username and password. This method connects and disconnects
        from the Venus OS device. Do not call connect() before this method."""

        try:
            await self.connect()
            return await self._read_installation_id()
        except MQTTConnectError as e:
            if "135" in str(e):
                raise InvalidAuthError from e
            raise CannotConnectError from e
        finally:
            await self.disconnect()

    async def _setup_subscriptions(self) -> None:
        """Subscribe to list of topics."""

        if self._client is None:
            raise ProgrammingError
        if not self._client.is_connected:
            raise NotConnectedError

        self._client.on_message = self._on_message

        for topic in topic_map:
            self._client.subscribe(topic)

        self._client.subscribe("N/+/full_publish_completed")

    async def _wait_for_first_refresh(self) -> None:
        """Wait for the first full refresh to complete, as per the
        "full_publish_completed" MQTT message."""
        await asyncio.wait_for(self._first_refresh_event.wait(), timeout=60)

    def _create_device_unique_id(self, installation_id: str, device_type: str, device_id: str) -> str:
        return f"{installation_id}_{device_type}_{device_id}"

    def _get_or_create_device(self, parsed_topic: ParsedTopic, desc: TopicDescriptor) -> Device:
        unique_id = self._create_device_unique_id(
            parsed_topic.installation_id,
            parsed_topic.device_type,
            parsed_topic.device_id,
        )
        device = self._devices.get(unique_id)
        if device is None:
            device = Device(
                unique_id,
                desc,
                parsed_topic.installation_id,
                parsed_topic.device_type,
                parsed_topic.device_id,
            )
            self._devices[unique_id] = device
            if parsed_topic.device_type == "system":
                if self.model_name is not None:
                    device.set_root_device_name(self._model_name)
                else:
                    device.set_root_device_name("Victron Venus")

        return device

    async def _on_message(
        self,
        client: gmqttClient,
        topic: str,
        payload: bytes,
        qos: int,
        retain: bool,  # noqa:  ARG002 pylint: disable=unused-argument
    ) -> None:
        """Handle an incoming message from the hub."""

        if "full_publish_completed" in topic:
            client.unsubscribe("N/+/full_publish_completed")
            self._first_refresh_event.set()
            return

        parsed_topic = ParsedTopic.from_topic(topic)
        if parsed_topic is None:
            return

        if parsed_topic.device_type != "system" and parsed_topic.device_id == "0":
            return

        desc = topic_map.get(parsed_topic.wildcards_with_device_type)
        if desc is None:
            desc = topic_map.get(parsed_topic.wildcards_without_device_type)

        if desc is None:
            return

        device = self._get_or_create_device(parsed_topic, desc)

        await device.handle_message(parsed_topic, desc, payload.decode())

    def get_device_from_unique_id(self, unique_id: str) -> Device:
        """Get a device from a unique id."""
        device = self._devices.get(unique_id)
        if device is None:
            raise KeyError(f"Device with unique id {unique_id} not found.")
        return device

    def _get_device_unique_id_from_metric_unique_id(self, unique_id: str) -> str:
        return "_".join(unique_id.split("_")[:3])

    def get_metric_from_unique_id(self, unique_id: str) -> Metric:
        """Get a metric from a unique id."""
        device = self.get_device_from_unique_id(self._get_device_unique_id_from_metric_unique_id(unique_id))
        return device.get_metric_from_unique_id(unique_id)

    @property
    def devices(self) -> list[Device]:
        "Return a list of devices attached to the hub. Requires initialize_devices_and_metrics() to be called first."
        return list(self._devices.values())

    @property
    def installation_id(self) -> str:
        """Return the installation id."""
        return self._installation_id

    @property
    def model_name(self) -> str:
        """Return the model name."""
        return self._model_name

    @property
    def connected(self) -> bool:
        """Return if connected."""
        if self._client is None:
            return False
        return self._client.is_connected


class CannotConnectError(Exception):
    """Error to indicate we cannot connect."""


class InvalidAuthError(Exception):
    """Error to indicate there is invalid auth."""


class ProgrammingError(Exception):
    """Error to indicate that we are in a state that should never be reached."""


class NotConnectedError(Exception):
    """Error to indicate that we expected to be connected at this stage but is not."""
